/*
 * sudo_server.c
 * Copyright (C) 2008 wantora
 * MIT License
 */

#include <windows.h>
#include <io.h>
#include <fcntl.h>
#include <stdio.h>
#include <tchar.h>
#include "sudo_server.h"
#include "utils.h"

static HANDLE hCEvent;

/* OtpCv̎M */
static void pipeReader(HANDLE hPipe, HANDLE hProcess) {
	char szBuf[1024];
	DWORD dwRead;
	DWORD dwWritten;
	DWORD dwAvail;
	HANDLE hStdout;
	
	hStdout = GetStdHandle(STD_OUTPUT_HANDLE);
	dwAvail = 0;
	while (dwAvail == 0) {
		if (!PeekNamedPipe(hPipe, NULL, 0, NULL, &dwAvail, NULL)
		      && WaitForSingleObject(hProcess, 10) == WAIT_OBJECT_0) return;
	}
	
	while (1) {
		WaitForSingleObject(hPipe, INFINITE);
		
		if (!ReadFile(hPipe, szBuf, sizeof(szBuf) - 1, &dwRead, NULL)) return;
		if (dwRead > 0) WriteFile(hStdout, szBuf, dwRead, &dwWritten, NULL);
	}
}

static int envsize(LPCTSTR env) {
	int len = 0;
	
	for (; env[len]; len++) {
		while (env[len]) len++;
	}
	
	return sizeof(*env) * (len + 1);
}

static void writeEnv(HANDLE hEnvPipe) {
	LPVOID env;
	DWORD dwWritten;
	
	env = GetEnvironmentStrings();
	
	ConnectNamedPipe(hEnvPipe, NULL);
	WriteFile(hEnvPipe, env, envsize(env), &dwWritten, NULL);
	
	FreeEnvironmentStrings(env);
}

/* Ctrl+C */
static BOOL WINAPI HandlerRoutine(DWORD dwCtrlType) {
	if (dwCtrlType == CTRL_C_EVENT || dwCtrlType == CTRL_CLOSE_EVENT) {
		SetEvent(hCEvent);
		return TRUE;
	}
	
	return FALSE;
}

/* M */
void serverMain(LPCTSTR sudo, LPCTSTR command, int mode) {
	LPTSTR lpParam;
	LPTSTR pipeName;
	HANDLE hPipe;
	HANDLE hEnvPipe;
	LPTSTR lpVerb;
	SHELLEXECUTEINFO si = {0};
	
	pipeName = getID();
	if (mode & SERVER_DEBUG_MODE) {
		lpVerb = TEXT("open");
	} else {
		lpVerb = TEXT("runas");
	}
	
	lpParam = (LPTSTR)malloc(sizeof(*lpParam) * (9+_tcslen(pipeName)+1+_tcslen(command)+1));
	_stprintf(lpParam, TEXT("--client %s %s"), pipeName, command);
	
	/*I*/
	hCEvent = CreateEvent(NULL, FALSE, FALSE, getCEventName(pipeName));
	if (hCEvent == NULL) return;
	SetConsoleCtrlHandler(HandlerRoutine, TRUE);
	
	/*pCv쐬*/
	hPipe = CreateNamedPipe(getPipeName(pipeName),
		PIPE_ACCESS_INBOUND, PIPE_TYPE_BYTE,
		1,  4095, 4095, 0, NULL);
	if (hPipe == INVALID_HANDLE_VALUE) return;
	
	hEnvPipe = CreateNamedPipe(getEnvPipeName(pipeName),
		PIPE_ACCESS_OUTBOUND, PIPE_TYPE_BYTE,
		1,  4095, 4095, 0, NULL);
	if (hEnvPipe == INVALID_HANDLE_VALUE) return;
	
	/*vZXN*/
	si.cbSize = sizeof(si);
	si.fMask = SEE_MASK_NOCLOSEPROCESS;
	si.lpVerb = lpVerb;
	si.lpFile = sudo;
	si.lpParameters = lpParam;
	if (mode & SERVER_DEBUG_MODE) {
		si.nShow = SW_SHOWNORMAL;
	} else {
		si.nShow = SW_HIDE;
	}
	
	if (ShellExecuteEx(&si)) {
		writeEnv(hEnvPipe);
		
		pipeReader(hPipe, si.hProcess);
		WaitForSingleObject(si.hProcess, INFINITE);
		
		CloseHandle(si.hProcess);
	}
	
	free(lpParam);
	CloseHandle(hPipe);
	CloseHandle(hEnvPipe);
	CloseHandle(hCEvent);
}
